/*
 *  Exception handler main initiation and exception handling.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

#define __EXC_INTERNAL
#include "exc.h"
#include "watch.h"
#include "printk.h"
#include "demangle.h"
#include "uasm.h"
#include "trace.h"

#define STATIC  /* for debugging */

#ifndef cdecl
#define cdecl
#endif

/*
 * Public data
 */
int    excAsmLines = 12;           /* # of lines to disassemble     */
int    excAsmDump  = 1;            /* do assembly dump at exception */
int    excSetText  = 1;            /* force textmode if graphics    */
int    excCoreFile = USECOREFILE;  /* dump text-messages to "core"  */
ULONG  excAppStart = 0;            /* application start (-offset x) */
ULONG  excAppEnd   = 0xFFFFFFFF;   /* application end offset        */
ULONG  excStkStart = APP_END();    /* lowest address for stack area */
ULONG  excStkEnd   = 0xFFFFFFFF;   /* highest     --- " ---         */

/*
 * Prototypes for locals
 */
STATIC UINT FindAppStart   (USHORT sel);
STATIC void FaultingFinish (void);
STATIC char ExcHasErrCode  (char i);

/*
 * Local data
 */
static char  videoMode   = 3;
static char  videoHeight = 24;
static void (*excHookUser)(excReg *r);

#define __EXC_C

#ifdef __DJGPP__
  #include "djgpp/exc_dj.c"
#else
  #include "pharlap/exc_ph.c"
#endif

/*
 * Routines common to Pharlap and djgpp targets
 */

char *AddressSymbol (USHORT cs, ULONG eip)
{
  static char buf [512];  /* C++ symbols are very long.. */
  char       *sym = GetSymbolStr (eip);
  int         len;
  if (cs)
       len = sprintk (buf, "%02X:%08lX", cs, eip);
  else len = sprintk (buf, "%08lX", eip);

  if (*sym)
     sprintk (buf+len," (%s)",sym);
  return (buf);
}

/*
 * Return TRUE if exception is known to have a
 * extra error-code on stack.
 */
STATIC char ExcHasErrCode (char i)
{
  return (i == INT_DBLF || (i >= INT_ITSS && i <= INT_PAGEF));
}
      
/*
 * Function used to figure out who called a zero-pointer.
 * i.e.
 *   foo = NULL;
 *    ..
 *   (*foo)();
 */
static ULONG GetCallee (ULONG esp)
{
  ULONG cs_limit;
  ULONG ds_limit;

  GetCodeLimit (&cs_limit);
  GetDataLimit (&ds_limit);

  if ((esp > excStkStart) && (esp < ds_limit))
  {
    ULONG addr = *(ULONG*)esp;
    if ((addr > excAppStart) && (addr < cs_limit) && (addr < excStkStart))
       return (addr);
  }
  return (0);
}

/*
 *  This is our high-level exception handler.
 *  For djgpp, this routine is called when signals (SIGSEGV,SIGFPE,
 *  SIGTRAP or SIGNOFP) are raised.
 *  For Pharlap, this routine is called from exc_0.asm when any
 *  exception is generated.
 */
STATIC void FaultingFinish (void)
{
  int    i;
  ULONG  callee = 0L;
  int    fd     = __exc_fd;
  char   *watch;
#ifdef __DJGPP__
  excReg *r = ExcFillContext (__djgpp_exception_state);
#else
  excReg *r = &context;
#endif

  static char reentry = 0;

  if (excStkStart < excAppStart)     /* excStkStart destroyed ! */
  {
    excStkStart = APP_END();
    fdprintk (STDERR_FILENO,
              "PANIC: exception handler data destroyed\r\n");
  }

  if (reentry)
  {
    fdprintk (STDERR_FILENO,
              "PANIC: exception handler reentered from %02X:%08X\r\n",
              r->CS, r->EIP);
    RemoveExcHandler();
    exc_exit (1);
  }

  if (excCoreFile)
  {
    fd = creat ("core", S_IWRITE);
    errno = 0;
  }

  reentry = 1;

  if (excSetText)     /* if graphics, set text mode */
     SetTextMode();

  switch (r->excNumber)
  {
#if 0
    case INT_DIV:
         r->EIP -= 2;
         break;
#endif
    case INT_INT3:
    case INT_INTO:
         r->EIP -= 1;
         break;

#if !defined(__DJGPP__)
    case 18: /* These are signals, hence no context */
    case 19:  
    case 20:
         fdprintk (fd, "\r\n%s\r\n", excName[r->excNumber]);
         return;
#endif
  }

  if (r->excNumber == INT_DBG &&
      (watch = GetDebugWatch(r->EIP)) != NULL)
       fdprintk (fd, "\r\nDebug watchpoint: %s\r\n", watch);
  else fdprintk (fd, "\r\n%s (exc %d) at %s",
                 excName[r->excNumber], r->excNumber,
                 AddressSymbol(r->CS, r->EIP));

  if ((r->EIP < excAppStart || r->EIP > excAppEnd) &&
      (callee = GetCallee(r->ESP)) != 0)
       fdprintk (fd, ". Entered from %s.\r\n", AddressSymbol(0,callee));
  else fdprintk (fd, ".\r\n");

  if (ExcHasErrCode(r->excNumber))
     fdprintk (fd, "Error code %02X.\r\n",r->excCode);

  fdprintk (fd,
            "Registers:\r\nEAX %08X  EBX %08X  ECX %08X  EDX %08X\r\n"
            "ESI %08X  EDI %08X  EBP %08X  ESP %08X\r\n"
            "DS  %04X      FS  %04X      SS  %04X      CR0 %08X\r\n"
            "ES  %04X      GS  %04X      FLG %08X  CR2 %08X\r\n",
            r->EAX,r->EBX,r->ECX,r->EDX,r->ESI,r->EDI,r->EBP,r->ESP,
            r->DS ,r->FS, r->SS ,r->CR0,r->ES, r->GS, r->EFL,r->CR2);

  if (excAsmDump && r->EIP >= excAppStart && r->EIP < excStkStart)
  {
    void *eip = (void*) r->EIP;
    void *adr = eip;
    
    fdprintk (fd, "\r\nInstruction trace:\r\n");
    for (i = 0; i < excAsmLines; i++)
    {
      const char *Asm = Disassemble (&adr);
      fdprintk (fd, "%02X:%08lX  %s\r\n", r->CS, (ULONG)eip, Asm);
      eip = adr;
    }
  }

#if STACKREWIND
  fdprintk (fd, "\r\nCall stack:\r\n");
  RewindStack (r->EIP, r->EBP, fd);
#endif

  if ((fd != STDOUT_FILENO) && (fd != STDERR_FILENO))
  {
    exc_close (fd);
    excCoreFile = 0;
    fdprintk (STDERR_FILENO, "Exception. Core dumped.\r\n");
  }

  RemoveExcHandler();

#if (USERHOOK) && (!REALLOC_SYM)
  if (excHookUser)
    (*excHookUser)(r);
#endif

#ifdef __HIGHC__
  /*
   * With stack checking on (`#pragma On(check_stack)'), HighC makes
   * this prologue in every function:
   *      push ebp
   *      mov  ebp,esp
   *      mov  ecx,esp
   *      sub  ecx,dword ptr _mwstack_limit+4
   *      cmp  ecx,local_stack_needed+12
   *      jge  $L1
   *      int  0
   * $L1: sub  esp,local_stack_needed
   *
   * We need to fix _mwstack_limit[1] in order not to get reentered in
   * any atexit() functions
   */
#if REALLOC_SYM          /* fix _mwstack_limit[1] in application DS */
  {
    ULONG _far *stack_start;
    FP_SET (stack_start, &_mwstack_limit[1], SS_DATA);
    *stack_start = (UINT)&exc_BOS; 
  }
#else
  _mwstack_limit[1] = (UINT)&exc_BOS;    /* set new Bottom-Of-Stack */
#endif

#endif  /* __HIGHC__ */

  reentry = 0;
  exc_exit (1);
}


