#include <stdarg.h>
#include <string.h>
#include <io.h>

#define __EXC_INTERNAL
#include "exc.h"
#include "printk.h"

#ifdef __DJGPP__
  #include <dos.h>
  #define exc_write(handle, buf, len)                \
           do {                                      \
             UINT result;                            \
             _dos_write (handle, buf, len, &result); \
           } while (0)
#endif

#define BUF_SIZE     512
#define is_digit(c)  ((c) >= '0' && (c) <= '9')

int __exc_fd = 1;  /* filehandle for stdout */


int printk (const char *fmt, ...)
{
  char    buf [BUF_SIZE];
  int     len;
  va_list args;

  va_start (args, fmt);
  len = vsprintk (buf, fmt, args);
  exc_write (__exc_fd, buf, len);
  va_end (args);
  return (len);
}

int fdprintk (int fd, const char *fmt, ...)
{
  char    buf [BUF_SIZE];
  int     len;
  va_list args;

  va_start (args, fmt);
  len = vsprintk (buf, fmt, args);
  exc_write (fd, buf, len);
  va_end (args);
  return (len);
}

int sprintk (char *buf, const char *fmt, ...)
{
  int     len;
  va_list args;

  va_start (args, fmt);
  len = vsprintk (buf, fmt, args);
  va_end (args);
  return (len);
}

#if !defined(__BORLANDC__) && !defined(__MSC32__)
  static size_t strlen (const char *str)
  {
    register const char *s;
    for (s = str; *s; ++s)
         ;
    return (s - str);
  }
#endif

static int skip_atoi (const char **s)
{
  int i = 0;

  while (is_digit(**s))
    i = i * 10 + *((*s)++) - '0';
  return (i);
}

#define ZEROPAD  1         /* pad with zero                    */
#define SIGN     2         /* unsigned/signed long             */
#define PLUS     4         /* show plus                        */
#define SPACE    8         /* space if plus                    */
#define LEFT    16         /* left justified                   */
#define SPECIAL 32         /* 0x                               */
#define SMALL   64         /* use 'abcdef' instead of 'ABCDEF' */


static char * number (char *str, long num, int base,
                      int size, int precision, int type)
{
  static char *digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  char   c, sign, tmp[36];
  int    i;

  if (type & SMALL)
      digits = "0123456789abcdefghijklmnopqrstuvwxyz";

  if (type & LEFT)
      type &= ~ZEROPAD;

  if (base < 2 || base > 36)
      return (0);

  if (type & ZEROPAD)
       c = '0';
  else c = ' ';

  if (type & SIGN && num < 0)
  {
    sign = '-';
    num  = -num;
  }
  else
    sign = (type & PLUS) ? '+' : ((type & SPACE) ? ' ' : 0);

  if (sign)
     size--;

  if (type & SPECIAL)
  {
    if (base == 16)
       size -= 2;
    else if (base == 8)
            size--;
  }
  i = 0;
  if (num == 0)
    tmp[i++] = '0';
  else
    while (num != 0)
    {
      ULONG _res = (ULONG) num % (ULONG) base;
      num = (ULONG) num / (ULONG) base;
      tmp[i++] = digits[_res];
    }

  if (i > precision)
     precision = i;
  size -= precision;
  if (!(type & (ZEROPAD + LEFT)))
     while (size-- > 0)
           *str++ = ' ';

  if (sign)
     *str++ = sign;
  if (type & SPECIAL)
  {
    if (base == 8)
        *str++ = '0';
    else if (base == 16)
    {
      *str++ = '0';
      *str++ = digits[33];
    }
  }

  if (!(type & LEFT))
     while (size-- > 0)
           *str++ = c;

  while (i < precision--)
        *str++ = '0';
  while (i-- > 0)
        *str++ = tmp[i];
  while (size-- > 0)
        *str++ = ' ';
  return (str);
}

int vsprintk (char *buf, const char *fmt, va_list args)
{
  char *str, *s = NULL;
  int  len = 0;
  int  i   = 0;
  int  use_long = 0;
  int  flags = 0;       /* flags to number()                         */
  int  field_width;     /* width of output field                     */
  int  precision = -1;  /* min. # of digits for integers; max number */
                        /* of chars for from string                  */

  for (str = buf; *fmt; ++fmt)
  {
    if (*fmt != '%')
    {
      *str++ = *fmt;
      continue;
    }
    /* process flags */
    flags = 0;

  repeat:
    ++fmt;                        /* this also skips first '%' */
    switch (*fmt)
    {
      case '-':
           flags |= LEFT;
           goto repeat;
      case '+':
           flags |= PLUS;
           goto repeat;
      case ' ':
           flags |= SPACE;
           goto repeat;
      case '#':
           flags |= SPECIAL;
           goto repeat;
      case '0':
           flags |= ZEROPAD;
           goto repeat;
      case '%':
           *str++ = '%';
           continue;
    }

    /* get field width
     */
    field_width = -1;
    if (is_digit(*fmt))
         field_width = skip_atoi (&fmt);
    else if (*fmt == '*')
    {
      /* it's the next argument
       */
      ++fmt; // !! added
      field_width = va_arg (args, int);
      if (field_width < 0)
      {
        field_width = -field_width;
        flags |= LEFT;
      }
    }
    /* get the precision
     */
    precision = -1;
    if (*fmt == '.')
    {
      ++fmt;
      if (is_digit(*fmt))
          precision = skip_atoi (&fmt);
      else if (*fmt == '*')
      {
        /* it's the next argument */
        precision = va_arg (args, int);
        ++fmt; // !! added
      }
      if (precision < 0)
          precision = 0;
    }
    if (*fmt == 'h' || *fmt == 'l' || *fmt == 'L')
    {
      if (*fmt != 'h')
         use_long = 1;
      ++fmt;
    }

    switch (*fmt)
    {
      case 'c':
           if (!(flags & LEFT))
              while (--field_width > 0)
                    *str++ = ' ';
           *str++ = (unsigned char) va_arg (args, int);
           while (--field_width > 0)
                 *str++ = ' ';
           break;

      case 's':
           s   = va_arg (args, char*);
           len = strlen (s);
           if (precision < 0)
               precision = len;
           else if (len > precision)
                    len = precision;

           if (!(flags & LEFT))
              while (len < field_width--)
                    *str++ = ' ';
           for (i = 0; i < len; ++i)
               *str++ = *s++;
           while (len < field_width--)
                 *str++ = ' ';
           break;

#ifdef not_used
      case 'o':
           str = number (str, va_arg(args, unsigned), 8,
                         field_width, precision, flags);
           break;

      case 'p':
           if (field_width == -1)
           {
             field_width = 8;
             flags |= ZEROPAD;
           }
           str = number (str, (unsigned)va_arg(args,void*), 16,
                         field_width, precision, flags);
           break;

      case 'n':
           {
             char *ip = (char*)va_arg(args, int *);
             *ip = (str - buf);
           }
           break;
#endif

      case 'x':
           flags |= SMALL;
           /* fall-through */

      case 'X':
           if (use_long)
                str = number (str, va_arg(args, ULONG), 16,
                              field_width, precision, flags);
           else str = number (str, va_arg(args, unsigned), 16,
                              field_width, precision, flags);
           break;

      case 'd':
      case 'i':
           flags |= SIGN;
           /* fall-through */

      case 'u':
           if (use_long)
                str = number (str, va_arg(args, unsigned), 10,
                              field_width, precision, flags);
           else str = number (str, va_arg(args, ULONG), 10,
                              field_width, precision, flags);
           break;

      default:
           if (*fmt != '%')
               *str++ = '%';
           if (*fmt)
                *str++ = *fmt;
           else fmt--;
           break;
    }
  }
  *str = '\0';
  return (str - buf);
}

