/*
 * This file contains code to print function/arg stack tracebacks at
 * exit time. Especially after an exception.  This code is derived from
 * the NOS (ka9q) networking package by Phil Karn - 1992
 *
 * 12-Sep 1995 - Gisle Vanem
 *   Rewritten to work with flat dos-extended applications using
 *   PharLap DOSX. Doesn't use .map file, but reads the .exe/.exp
 *   symbol table at init-time
 *
 * 18-Nov 1998 - Gisle Vanem
 *   Added support for djgpp 2.x and COFF/AOUT object formats
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <fcntl.h>
#include <io.h>

/* r-mode test program
 */
#if (defined(__SMALL__) && !defined(DOS386)) || defined(__LARGE__)
  #define FARPTR char far*
  #include "trace.h"
  #include "demangle.h"
  #include <plexp.h>
  #define _dx_hdroff_get(h,o,e,d) (1==0)
  #pragma warn -unr

#else
  #define __EXC_INTERNAL
  #include "exc.h"
  #include "trace.h"
  #include "printk.h"
  #include "demangle.h"
#endif

#ifdef __WATCOMC__
  #undef  FP_OFF
  #define FP_OFF(p) ((unsigned)(p))
#endif

#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

#ifndef cdecl
#define cdecl
#endif

#define STATIC

typedef int (*CmpFunc)(const void*, const void*);

STATIC char  *LoadSymTable      (int handle);
STATIC int    CompareName       (const SymNode **a, const SymNode **b);
STATIC USHORT GetSelectorVal    (USHORT idx);
STATIC int    SortSymbolsByAddr (unsigned num);

int    excDemangleCpp = DEMANGLE_CPP; /* demangle C++ symbols */
int    excCplusPlusSyms;
int    excProgIsCplusPlus;

static SymNode *__syms;
static SegNode  segTable [MAX_SEL];

static ULONG  numPublics   = 0;
static ULONG  symTotalSize = 0;
static char  *symTooBig    = "Symbol table too big";

/*
 * Try to initialize stack traceback handling.
 * Get public symbols from executable image.
 */
char *InitStackTrace (char *name)
{
  int  handle;
  char *err;

  if (!name)
     return ("No file");

#ifdef __DJGPP__
  _fmode = O_BINARY;
  handle = open (name, O_RDONLY | O_BINARY);
  if (handle <= 0)
     return ("Cannot open AOUT/COFF-file");
#else
  handle = open (name, O_RDONLY | O_BINARY);
  if (handle <= 0)
     return ("Cannot open exp/exe-file");
#endif

  __syms = NULL;
  excCplusPlusSyms   = 0;
  excProgIsCplusPlus = 0;
  err = LoadSymTable (handle);
  close (handle);

  if (!err && !SortSymbolsByAddr(numPublics))
     err = symTooBig;

#if !defined(__DJGPP__)
  if (err)
  {
    /* We encountered problems loading symbols from executable.
     * See if a .sym file is present in same directory as .exe-file.
     * Build linked list from that file instead.
     */
    char  fname  [256];
    char  buffer [100];
    ULONG num = 0;
    FILE  *fil;

    strcpy (fname, name);
    memcpy (strrchr(fname,'.'), ".sym", 5);
    fil = fopen (fname, "rt");
    if (!fil)
       return (err);

    while (fgets(buffer,sizeof(buffer)-1,fil))
    {
      ULONG ofs = 0;
      char *cp, *symName;

      if (strnicmp(buffer,"-DEFINE ",8))  /* "-DEFINE %s=%X08H\n" */
         break;
      cp = buffer + 8;
      symName = cp;
      cp = strchr (cp, '=');
      if (!cp)
         continue;
      *cp++ = 0;
      if (sscanf(cp,"%lX",&ofs) != 1)
         continue;

      if (!AddAddrSymbol(symName,ofs,SS_CODE))
         break;
      num++;
    }
    fclose (fil);
    numPublics = num;
    err = (SortSymbolsByAddr(num) ? NULL : symTooBig);
  }
#endif

  return (err);
}

/*
 * Return total size of SymNode list and number of symbols
 */  
ULONG GetSymbolsSizes (ULONG *num)
{
  if (num)
     *num = numPublics;
  return (symTotalSize);
}

#if !defined(TEST)

/*
 * Do a stack rewind to see what functions where 
 * called with which arguments
 */
void RewindStack (ULONG start, ULONG base, int fd)
{
  int    loop    = 10;            /* print max 10 stack frames  */
  int    width   = excProgIsCplusPlus ? -50 : -20;
  ULONG  eip     = start;
  ULONG *context = (ULONG*) base; /* EBP increases as we rewind */
  ULONG  limit;

  GetCodeLimit (&limit);

  while (context >= (ULONG*)base && ((ULONG)context & 3) == 0 &&
         context >= (ULONG*)excStkStart && --loop)
  {
    ULONG   *next;
    SymNode *sp;

    if ((ULONG)context >= limit - 4   || eip >= limit ||
        (ULONG)context <= excAppStart || eip <= excAppStart)
       break;

    if ((sp = FindPublicSym(eip)) != NULL)
    {
      char  buf [MAX_SYM_NAME+10];
      ULONG diff = eip - sp->ofs;

      if (diff)
      {
        sprintk (buf, "%s+0%lX", sp->name, diff);
        fdprintk (fd, "%*s ", width, buf);
      }
      else
      {
        width = min (10, width);
        fdprintk (fd, "%*s ", width, sp->name);
      }
    }
    else
      fdprintk (fd, "%08lX %*s", eip, width-8, "");

    next = (ULONG*)*context;          /* get ESP stack value from context */

    /* Compute number of args to display
     */
    if (next == NULL ||               /* we're rewound back to CRT */
        next <= (ULONG*)excAppStart)
       fdprintk (fd, "( *** )");

    else
    {
      ULONG *argv;
      ULONG  argc = next - context - 2;

      argc = (argc > 10) ? 10 : argc;
      fdprintk (fd, "(");
    
      argv = context + 2; /* Args start after saved EBP and return address */
                          /* If Pascal style, args are in reverse order    */
      while (argc--)
      {
        if (argc == 0)
             fdprintk (fd, "%04lX)\r\n", *argv++);
        else fdprintk (fd, "%04lX ",     *argv++);
      }
    }

    eip     = *(context+1);  /* get next return address */
    context = next;
  }
  fdprintk (fd, "\r\n");
}


/*
 * Find the closest (lower) public symbol to ofs
 */       
SymNode * FindPublicSym (ULONG ofs)
{
  SymNode *sp, *prev = NULL;

  for (sp = __syms; sp; prev = sp, sp = sp->next)
  {
    if (ofs == sp->ofs)
       return (sp);
    if (sp->ofs > ofs)
       break;
  }
  return (prev);
}

/*
 * Return string for symbol at EIP
 */
char * GetSymbolStr (ULONG eip)
{
  SymNode *sp = FindPublicSym (eip);

  if (sp)
  {
    static char buf[MAX_SYM_NAME+10];
    ULONG  delta = eip - sp->ofs;

    if (delta)
         sprintk (buf, "%.*s+%lX",MAX_SYM_NAME, sp->name, delta);
    else sprintk (buf, "%.*s",    MAX_SYM_NAME, sp->name);
    return (buf);
  }
  return ("");
}


/*
 * Move the public symbol tables to a buffer in another segment.
 * Free allocated memory along the way. Only for compilers that
 * have 48-bit far-pointers
 */
#if (__CMPLR_FEATURES__ & __FEATURE_FARPTR__)

unsigned ReallocPubSyms (FARPTR symArea, ULONG max)
{
  SymNode *sp, *sp0, *next;
  ULONG   total = 0L;

  if (max < symTotalSize)
     return (0);

  sp0 = NULL;
  numPublics = 0;

  for (sp = __syms; sp; sp = next)
  { 
    SymNode _far *new = (SymNode _far*) symArea;
    ULONG   size = sizeof(SymNode) + strlen(sp->name) + 1;

    if ((total += size) > max)
       break;
    new->sel  = sp->sel;
    new->ofs  = sp->ofs;
    new->next = sp0;
    /*
     * put symbol name after new node
     */
    far_memcpy (new+1, &sp->name, size-sizeof(SymNode));
    next = sp->next;
    free (sp->name);
    free (sp);
    sp0 = (SymNode*) FP_OFF (new);
    numPublics++;
    FP_INCR (symArea, size);
  }                       
  __syms = sp0;    /* updated lists start in new selector */
  return (total);
}

#else
unsigned ReallocPubSyms (FARPTR symArea, ULONG max)
{
  ARGSUSED (symArea);
  ARGSUSED (max);
  return (0);
}
#endif

#endif  /* !defined (TEST) */


#if defined(__DJGPP__)

static ULONG ProcessFile (int fd, long ofs)
{
  short sign;
  short size[2];

  if (ofs > 0)
     lseek (fd, ofs, SEEK_SET);
  read (fd, &sign, sizeof(sign));

  switch (sign)
  {
    case SIGN_MZ:        /* .exe */
         read (fd,&size,sizeof(size));
         ofs  = size[1] << 9;          /* # of 512 blocks */
         ofs -= size[0] ? 512 : 0;
         ofs += size[0];               /* total stub size */
         return ProcessFile (fd, ofs);

    case I386MAGIC:        /* .coff */
         return GetCoffSymbols (fd, ofs);

    case 0x010b:	/* a.out ZMAGIC */
    case 0x0107:	/* a.out object */
         return GetAoutSymbols (fd, ofs);
  }
  fprintf (stderr,"Unknown signature %04X\n", sign);
  return (0);
}

/*
 * Read symbols from .exe file.
 * Check headers and load public symbols.
 */
STATIC char * LoadSymTable (int handle)
{
  int i;

  memset (&segTable,0,sizeof(segTable));
  for (i = 0; i < MAX_SEL; i++)
     segTable[i].name = "?";

  numPublics = ProcessFile (handle,0);

  if (!numPublics)
     return ("Unknown COFF/AOUT format");
  return (NULL);
}

#else

/*
 * Read symbols from .exp/.exe file. Might be a bound exe-file.
 * Check headers and load public symbols (Pharlap or Codeview).
 */
STATIC char * LoadSymTable (int handle)
{
  EXP_HDR   hdr;
  OEXP_HDR *ehdr;       /* DOS .exe header = old .exp header */
  ULONG     num, size;
  ULONG     ofs = 0;
  SymHeader sh;
  UCHAR     codeView  = 0;
  UCHAR     pharLap   = 0;
  char     *buf;

  read (handle,(char*)&hdr,sizeof(hdr));

  if (hdr.exp_sign != SIGN_P3)           /* no "P3" signature */
  {
    UINT err, dosErr;
    ehdr = (OEXP_HDR*)&hdr;              /* see if exp file has a stub */

    if (ehdr->exe_sign != SIGN_MZ &&     /* "MZ" signature             */
        ehdr->exe_sign != SIGN_MQ)       /* "MQ" signature (REX file)  */
       return ("Not an executable file");

#if 0   // !! to-do
    if (ehdr->exe_sign == SIGN_MQ)
       rexFormat = 1;
#endif

    /* get bound exe-file EXP offset
     */
    if (_dx_hdroff_get(handle,&ofs,&err,&dosErr))
    {
      ofs  = ehdr->exe_size << 9;          /* # of 512 blocks */
      ofs -= ehdr->exe_szrem ? 512 : 0;
      ofs += ehdr->exe_szrem;              /* total stub size */
    }
    lseek (handle,ofs,SEEK_SET);
    read (handle,(char*)&hdr,sizeof(hdr));
  }

  if (hdr.exp_sign == SIGN_P2)             /* "P2" signature ? */
     return ("Simple EXP not supported");

  if (hdr.exp_sign != SIGN_P3)             /* "P3" signature ? */
     return ("Unknown EXP/REX format");

  if (hdr.exp_sym == 0L)
     return ("No symbol table found");

  lseek (handle,ofs+hdr.exp_sym,SEEK_SET); /* seek to symbol header  */
  read (handle,(char*)&sh,sizeof(sh));     /* read it                */

  codeView = (strncmp(sh.name,"SYM2",4) == 0); /* MS CodeView style  */
  pharLap  = (strncmp(sh.name,"SYM1",4) == 0); /* PharLap -sym style */

  if (!codeView && !pharLap)
     return ("Unknown symbol format");

  if (codeView)
  {
#if 1
    return ("CodeView not supported");
#else
    size = sh.pubSymSize;
#endif
  }
  else
    size = sh.pubSymSize + sh.segSymSize;   /* total symbol size */

  if ((buf = malloc(size)) == NULL)
     return (symTooBig);

  read (handle,buf,size);        /* read segment and public symbols */
  memset (&segTable,0,sizeof(segTable));

  if (codeView)
       num = GetCodeViewSymbols(buf,&sh);
  else num = GetPharLapSymbols (buf,&sh);

  free (buf);
  numPublics = num;
  return (NULL);
}
#endif  /* !__DJGPP__ */


/*
 * Demangle i.e. make the C++ symbols humanly readable
 * Unfortunately Metaware uses a proprietary mangling format.
 */
#if !defined(__HIGHC__)
static char * DemangleSymbol (char *name)
{
  char *new_name;

#ifdef __GNUC__
  int skip = 0;

  /*
   * GNU C++ symbols may start with '.'
   */
  if (name[0] == '.')
     skip++;
  new_name = cplus_demangle (name+skip, DMGL_PARAMS|DMGL_ANSI|DMGL_GNU);

#else
  /*
   * Non-GNU C++ (ARM) symbols always (?) contain '@'
   */
  if (!strchr(name,'@'))
     return (name);

  new_name = cplus_demangle (name, DMGL_PARAMS|DMGL_ANSI|DMGL_ARM);
#endif

  if (!new_name)
     return (name);

  excCplusPlusSyms++;
  excProgIsCplusPlus = 1;
  return (new_name);
}
#endif


/*
 * Check symbol name for non-alphas, demangle C++ and allocate
 * string for symbol and add to linked list.
 */
SymNode *AddAddrSymbol (char *name, ULONG ofs, USHORT sel)
{
  SymNode *sp;

  /* COFF symbols may contain non-alpha characters
   */
#ifdef __GNUC__
  char buf [MAX_SYM_NAME];
  int  i = 0;
  int  c = *name++;

  while (c && (isalnum(c) || c == '_' || c == '$' || c == '.'))
  {
    if (i >= MAX_SYM_NAME-1)
       break;
    buf[i++] = c;
    c = *name++;
  }
  buf[i] = '\0';
  name   = buf;
  if (i == 0)
     return (NULL);

#endif

  sp = calloc (sizeof(*sp),1);
  if (!sp)
     return (NULL);

#if !defined(__HIGHC__)
  if (excDemangleCpp)
  {
    char *cpp_name = DemangleSymbol (name);
    if (cpp_name)
       name = cpp_name;
  }
#endif

  sp->name = strdup (name);
  if (!sp->name)
     return (NULL);

  sp->ofs  = ofs;
  sp->sel  = sel;
  sp->next = __syms;
  __syms   = sp;
  symTotalSize += sizeof(*sp) + strlen(name) + 1;
  return (sp);
}

/*
 * Allocate selector name to segtable[] array.
 */
SegNode *AddSegSymbol (char *name, ULONG ofs, ULONG sel, ULONG limit)
{
  static int idx = 0;
  SegNode   *st;

  if (idx < 0 || idx >= MAX_SEL)
     return (NULL);

  st = &segTable [idx++];
  st->name = strdup (name);
  if (!st->name)
     return (NULL);

  st->ofs      = ofs;
  st->selector = sel;
  st->limit    = limit;
  return (st);
}

/*
 * Sort the public symbols by address (ascending order)
 */
STATIC int SortSymbolsByAddr (unsigned num)
{
  int i;
  SymNode  *sp;
  SymNode **spp = malloc (num * sizeof(SymNode*));

  if (spp == NULL)
     return (0);

  for (i = 0, sp = __syms; sp; i++, sp = sp->next)
      spp[i] = sp;

  qsort (spp, num, sizeof(SymNode*), (CmpFunc)CompareName);
  __syms = NULL;

  for (i = num-1; i >= 0; i--)  /* Put them back in linked list */
  {
    sp       = spp[i];
    sp->next = __syms;
    __syms   = sp;
  }
  free (spp);
  return (1);
}

/*
 * Return selector value from selector index
 */
USHORT GetSelectorVal (USHORT idx)
{
  if (!idx || idx >= MAX_SEL)
     return (0);
  return (segTable[idx-1].selector);
}
       
/*
 * Comparision routine needed by Quick-sort routine
 */
STATIC int CompareName (const SymNode **a, const SymNode **b)
{
  if ((*a)->sel > (*b)->sel) return (+1);
  if ((*a)->sel < (*b)->sel) return (-1);
  if ((*a)->ofs > (*b)->ofs) return (+1);
  if ((*a)->ofs < (*b)->ofs) return (-1);
  return strcmp ((*a)->name,(*b)->name);
}


/*
 * Functions used by test programs EXE_DMP or EXP_DMP.
 */
#if defined(TEST)

static char * GetSelectorName (USHORT sel)
{
  int i;
  for (i = 0; i < MAX_SEL; i++)
    if (segTable[i].selector == sel)
       return (segTable[i].name);
  return (NULL);
}

static void ShowPublicSyms (void)
{
  SymNode *sp;
  int  width = excProgIsCplusPlus ? 50 : 32;

  printf ("Symbol %*sOffset    Selector\n", width-6,"");
  for (sp = __syms; sp; sp = sp->next)
      printf ("%-*s %08lX  %02X  %s\n",
              width, sp->name, sp->ofs, sp->sel, GetSelectorName(sp->sel));
}

#ifndef __DJGPP__
static void ShowPublicSegs (void)
{
  int i;

  puts ("\n\nSelector info:\nName\t   Sel  Offset   Limit\n");
  for (i = 0; i < MAX_SEL; i++)
     printf ("%-10s %02X   %08lX %08lX\n",
             segTable[i].name,segTable[i].selector,
             segTable[i].ofs, segTable[i].limit);
}
#endif

static void ShowSummary (void)
{
  printf ("Summary: %d C-symbols, %d C++ symbols\n"
          "        %d bytes allocated for symbols\n",
          numPublics - excCplusPlusSyms, excCplusPlusSyms, symTotalSize);
}

int main (int argc, char **argv)
{
  char *err;

  if (argc != 2)
  {
#ifdef __DJGPP__
    printf ("Usage:\n  %s Plain COFF/AOUT or stubbed file",argv[0]);
#else
    printf ("Usage:\n  %s PharLap exe/exp/rex-file",argv[0]);
#endif
    return (1);
  }
  err = InitStackTrace (argv[1]);
  if (err)
  {
    printf ("Cannot read symbols: %s\n",err);
    return (1);
  }

  ShowPublicSyms();
#ifndef __DJGPP__
  ShowPublicSegs();
#endif

  ShowSummary();
  return (0);
}

#endif  /* TEST */


