/*
 *  MMURTL Operating System Source Code
 *  Copyright 1991,1992,1993, Richard A. Burgess
 *  ALL RIGHTS RESERVED
 *  Version x0.8
 *
 *  Modification by G.Vanem <giva@bgnett.no>:
 *    Added decoding of FPU, MMX and Pentium Pro instructions
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#define __EXC_INTERNAL
#include "exc.h"
#include "uasm.h"
#include "printk.h"
#include "trace.h"

static void Decode  (char *s);
static void do_ModRM(char t);

/*
  The Intel 386 DX Programmer's Reference Manual provides a table that
  uses the following codes to assist in disassembly of 386 code (page A-3).
  The letters below are the same as the codes in the manual.  The ~ (tilde)
  is an escape character to signify expansion of the codes is
  required when the string is being output.

  Tilde tokens in strings:

     First char after '~':
   A - Direct address
   C - Reg of R/M picks control register
   D - Reg of R/M picks debug register
   E - R/M picks operand
   F - Flag register
   G - Reg of R/M selects a general register
   I - Immediate data
   J - Relative IP offset. don't get name for IP
   j - Relative IP offset. print IP as name+deltaIP
   M - R/M picks memory
   O - No R/M, offset only
   R - Mod of R/M picks register only
   S - Reg of R/M picks segment register
   T - Reg of R/M picks test register
   X - DS:ESI
   Y - ES:EDI
   Z - Sink; skip imm byte (divisor) if equal 10 (for AAD/AAM)
   2 - prefix of two-byte opcode
   e - put in 'E' if use32 (second char is part of reg name)
       put in 'W' if use16 or 'D' for use32 (second char is 'W')
   f - Floating point (second char is ESC value)
   g - do R/M group 'n'
   p - prefix
   s - Size override (second char is a,o)

     Second char after '~':
   a - Two words in memory (BOUND)
   b - Byte
   c - Byte or word
   d - DWord
   q - QWord
   p - 32 or 48 bit pointer
   s - Six byte pseudo-descriptor
   v - Word or Dword
   w - Word
   x - Put in 'E' in instruction if opsize = 32 (JECXZ)
   F - use floating reg in mod/rm
   M - use MMX reg in mod/rm
   X - use XMM reg in mod/rm
   1-9,A-C - group number, ESC value, etc
*/

static char *opmap1[256] = {

/* 0x */
  "ADD ~Eb,~Gb",    "ADD ~Ev,~Gv",      "ADD ~Gb,~Eb",   "ADD ~Gv,~Ev",
  "ADD AL,~Ib",     "ADD ~eAX,~Iv",     "PUSH ES",       "POP ES",
  "OR ~Eb,~Gb",     "OR ~Ev,~Gv",       "OR ~Gb,~Eb",    "OR ~Gv,~Ev",
  "OR AL,~Ib",      "OR ~eAX,~Iv",      "PUSH CS",       "~2 ",

/* 1x */
  "ADC ~Eb,~Gb",    "ADC ~Ev,~Gv",      "ADC ~Gb,~Eb",   "ADC ~Gv,~Ev",
  "ADC AL,~Ib",     "ADC ~eAX,~Iv",     "PUSH SS",       "POP SS",
  "SBB ~Eb,~Gb",    "SBB ~Ev,~Gv",      "SBB ~Gb,~Eb",   "SBB ~Gv,~Ev",
  "SBB AL,~Ib",     "SBB ~eAX,~Iv",     "PUSH DS",       "POP DS",

/* 2x */
  "AND ~Eb,~Gb",    "AND ~Ev,~Gv",      "AND ~Gb,~Eb",   "AND ~Gv,~Ev",
  "AND AL,~Ib",     "AND ~eAX,~Iv",     "~pE",           "DAA",
  "SUB ~Eb,~Gb",    "SUB ~Ev,~Gv",      "SUB ~Gb,~Eb",   "SUB ~Gv,~Ev",
  "SUB AL,~Ib",     "SUB ~eAX,~Iv",     "~pC",           "DAS",

/* 3x */
  "XOR ~Eb,~Gb",    "XOR ~Ev,~Gv",      "XOR ~Gb,~Eb",   "XOR ~Gv,~Ev",
  "XOR AL,~Ib",     "XOR ~eAX,~Iv",     "~pS",           "AAA",
  "CMP ~Eb,~Gb",    "CMP ~Ev,~Gv",      "CMP ~Gb,~Eb",   "CMP ~Gv,~Ev",
  "CMP AL,~Ib",     "CMP ~eAX,~Iv",     "~pD",           "AAS",

/* 4x */
  "INC ~eAX",       "INC ~eCX",         "INC ~eDX",      "INC ~eBX",
  "INC ~eSP",       "INC ~eBP",         "INC ~eSI",      "INC ~eDI",
  "DEC ~eAX",       "DEC ~eCX",         "DEC ~eDX",      "DEC ~eBX",
  "DEC ~eSP",       "DEC ~eBP",         "DEC ~eSI",      "DEC ~eDI",

/* 5x */
  "PUSH ~eAX",      "PUSH ~eCX",        "PUSH ~eDX",     "PUSH ~eBX",
  "PUSH ~eSP",      "PUSH ~eBP",        "PUSH ~eSI",     "PUSH ~eDI",
  "POP ~eAX",       "POP ~eCX",         "POP ~eDX",      "POP ~eBX",
  "POP ~eSP",       "POP ~eBP",         "POP ~eSI",      "POP ~eDI",

/* 6x */
  "PUSHAD",         "POPAD",            "BOUND ~Gv,~Ma", "ARPL ~Ew,~Rw",
  "~pF",            "~pG",              "~so",           "~sa",
  "PUSH ~Iv",       "IMUL ~Gv,~Ev*~Iv", "PUSH ~Ib",      "IMUL ~Gv,~Ev*~Ib",
  "INSB ~Yb,DX",    "INS~eW ~Yv,DX",    "OUTSB DX,~Xb",  "OUTS~eW DX,~Xv",

/* 7x */
  "JO ~Jb",         "JNO ~Jb",          "JNC ~Jb",       "JC ~Jb",
  "JZ ~Jb",         "JNZ ~Jb",          "JBE ~Jb",       "JNBE ~Jb",
  "JS ~Jb",         "JNS ~Jb",          "JPE ~Jb",       "JPO ~Jb",
  "JL ~Jb",         "JGE ~Jb",          "JLE ~Jb",       "JG ~Jb",

/* 8x */
  "~g1 ~Eb,~Ib",    "~g1 ~Ev,~Iv",      "MOV AL,~Ib",    "~g1 ~Ev,~Ib",
  "TEST ~Eb,~Gb",   "TEST ~Ev,~Gv",     "XCHG ~Eb,~Gb",  "XCHG ~Ev,~Gv",
  "MOV ~Eb,~Gb",    "MOV ~Ev,~Gv",      "MOV ~Gb,~Eb",   "MOV ~Gv,~Ev",
  "MOV ~Ew,~Sw",    "LEA ~Gv,~M*",      "MOV ~Sw,~Ew",   "POP ~Ev",

/* 9x */
  "NOP",            "XCHG ~eAX,~eCX",   "XCHG ~eAX,~eDX","XCHG ~eAX,~eBX",
  "XCHG ~eAX,~eSP", "XCHG ~eAX,~eBP",   "XCHG ~eAX,~eSI","XCHG ~eAX,~eDI",
  "CBW",            "CDW",              "CALL ~Ap",      "FWAIT",
  "PUSH ~eFLAGS",   "POP ~eFLAGS",      "SAHF",          "LAHF",

/* Ax */
  "MOV AL,~Ov",     "MOV ~eAX,~Ov",     "MOV ~Ov,AL",    "MOV ~Ov,~eAX",
  "MOVSB",          "MOVS~eW",          "CMPSB ~Xb,~Yb", "CMPS~eW ~Xv,~Yv",
  "TEST AL,~Ib",    "TEST ~eAX,~Iv",    "STOSB ~Yb,AL",  "STOS~eW",
  "LODSB",          "LODS~eW",          "SCASB",         "SCAS~eW",

/* Bx */
  "MOV AL,~Ib",     "MOV CL,~Ib",       "MOV DL,~Ib",    "MOV BL,~Ib",
  "MOV AH,~Ib",     "MOV CH,~Ib",       "MOV DH,~Ib",    "MOV BH,~Ib",
  "MOV ~eAX,~Iv",   "MOV ~eCX,~Iv",     "MOV ~eDX,~Iv",  "MOV ~eBX,~Iv",
  "MOV ~eSP,~Iv",   "MOV ~eBP,~Iv",     "MOV ~eSI,~Iv",  "MOV ~eDI,~Iv",

/* Cx */
  "~g2 ~Eb,~Ib",    "~g2 ~Ev,~Ib",      "RET ~Iw",       "RET",
  "LES ~Gv,~Mp",    "LDS ~Gv,~Mp",      "MOV ~Eb,~Ib",   "MOV ~Ev,~Iv",
  "ENTER ~Iw,~Ib",  "LEAVE",            "RETF ~Iw",      "RETF",
  "INT 3",          "INT ~Ib",          "INTO",          "IRETD",

/* Dx */
  "~g2 ~Eb,1",      "~g2 ~Ev,1",        "~g2 ~Eb,CL",    "~g2 ~Ev,CL",
  "AAM ~Z",         "AAD ~Z",           "SALC" ,         "XLAT",
  "~f0",            "~f1",              "~f2",           "~f3",
  "~f4",            "~f5",              "~f6",           "~f7",

/* Ex */
  "LOOPNE ~Jb",     "LOOPE ~Jb",        "LOOP ~Jb",      "J~xCXZ ~Jb",
  "IN AL,~Ib",      "IN ~eAX,~Ib",      "OUT ~Ib,AL",    "OUT ~Ib,~eAX",
  "CALL ~jv",       "JMP ~Jv",          "JMP ~Ap",       "JMP ~Jb",
  "IN AL,DX",       "IN ~eAX,DX",       "OUT DX,AL",     "OUT DX,~eAX",

/* Fx */
  "LOCK: ~p*",      "ICEBP",            "REPNE ~p*",     "REP ~p*",
  "HLT",            "CMC",              "~g3",           "~g0",
  "CLC",            "STC",              "CLI",           "STI",
  "CLD",            "STD",              "~g4",           "~g5"
};


static char *secOpcode[256] = {   /* OP-code map for instructions 0F xx */

/* 0x */
  "~g6",               "~g7",               "LAR ~Gv,~Ew",       "LSL ~Gv,~Ew",
  "LOADALL",           "LOADALL",           "CLTS",              "LOADALL",
  "INVD",              "WBINVD",            0,                   "UD2",
  0,                   0,                   0,                   0,

/* 1x */
  "UMOV ~Eb,~Gb",      "UMOV ??",           "UMOV ??",           "UMOV ??",
  0,                   0,                   0,                   0,
  0,                   0,                   0,                   0,
  0,                   0,                   0,                   0,

/* 2x */
  "MOV ~Rd,~Cd",       "MOV ~Rd,~Dd",       "MOV ~Cd,~Rd",       "MOV ~Dd,~Rd",
  "MOV ~Rd,~Td",       0,                   "MOV ~Td,~Rd",       0,
  0,                   0,                   "CVTPI2PS ~GX,??",   0,
  "CVTPS2PI ~GM,??",   "CVTTPS2PI ~GM,??",  0,                   0,

/* 3x */
  "WRMSR",             "RDTSC",             "RDMSR",             "RDPMC",
  0,                   0,                   0,                   0,
  0,                   0,                   0,                   0,
  0,                   0,                   0,                   0,

/* 4x */
  "CMOVO ~Gv,~Ev",     "CMOVNO ~Gv,~Ev",    "CMOVB ~Gv,~Ev",     "CMOVNB ~Gv,~Ev",
  "CMOVE ~Gv,~Ev",     "CMOVNE ~Gv,~Ev",    "CMOVBE ~Gv,~Ev",    "CMOVA ~Gv,~Ev",
  "CMOVS ~Gv,~Ev",     "CMOVNS ~Gv,~Ev",    "CMOVP ~Gv,~Ev",     "CMOVNP ~Gv,~Ev",
  "CMOVL ~Gv,~Ev",     "CMOVNL ~Gv,~Ev",    "CMOVLE ~Gv,~Ev",    "CMOVG ~Gv,~Ev",

/* 5x */
  0,                   0,                   0,                   0,
  "ANDPS ~GX,??",      "ANDPS ~GX,??",      0,                   0,
  "ADDPS ~GX,??",      0,                   0,                   0,
  0,                   "MINPS ~GX,??",      "DIVPS ~GX,??",      "MAXPS ~GX,??",

/* 6x */
  "PUNPCKLBW ~GM,~Eq", "PUNPCKLWD ~GM,~Eq", "PUNPCKLDQ ~GM,~Eq", "PACKSSWB ~GM,~Eq",
  "PCMPGTB ~GM,~Eq",   "PCMPGTW ~GM,~Eq",   "PCMPGTD ~GM,~Eq",   "PACKUSWB ~GM,~Eq",
  "PUNPCKHBW ~GM,~Eq", "PUNPCKHWD ~GM,~Eq", "PUNPCKHDQ ~GM,~Eq", "PACKSSDW ~GM,~Eq",
  0,                   0,                   "MOVD ~GM,~Ed",      "MOVQ ~GM,~Eq",

/* 7x */
  0,                   "~g9 ~GM,~Ib",       "~g9 ~GM,~Ib",       "~g9 ~GM,~Ib",
  "PCMPEQB ~GM, ~Eq",  "PCMPEQW ~GM,~Eq",   "PCMPEQD ~GM,~Eq",   "EMMS",
  "SVDC ??",           "RSDC ??",           "SVLDT ??",          "RSLDT ??",
  "SVTS ??",           "RSTS ??",           "MOVD ~Ed,~GM",      "MOVQ ~Eq,~GM",

/* 8x */
  "JO ~Jv",            "JNO ~Jv",           "JC ~Jv",            "JNC ~Jv",
  "JZ ~Jv",            "JNZ ~Jv",           "JBE ~Jv",           "JNBE ~Jv",
  "JS ~Jv",            "JNS ~Jv",           "JPE ~Jv",           "JPO ~Jv",
  "JL ~Jv",            "JGE ~Jv",           "JLE ~Jv",           "JG ~Jv",

/* 9x */
  "SETO ~Eb",          "SETNO ~Eb",         "SETNC ~Eb",         "SETC ~Eb",
  "SETZ ~Eb",          "SETNZ ~Eb",         "SETBE ~Eb",         "SETNBE ~Eb",
  "SETS ~Eb",          "SETNS ~Eb",         "SETP ~Eb",          "SETNP ~Eb",
  "SETL ~Eb",          "SETGE ~Eb",         "SETLE ~Eb",         "SETG ~Eb",

/* Ax */
  "PUSH FS",           "POP FS",            "CPUID",             "BT ~Ev,~Gv",
  "SHLD ~Ev,~Gv,~Ib",  "SHLD ~Ev,~Gv,CL",   "XBTS ~Eb,~Gb",      "IBTS ~Ev,~Gv",
  "PUSH GS",           "POP GS",            "RSM",               "BTS ~Ev,~Gv",
  "SHRD ~Ev,~Gv,~Ib",  "SHRD ~Ev,~Gv,CL",   0,                   "IMUL ~Gv,~Ev",

/* Bx */
  "CMPXCHG ~Eb,~Gb",   "CMPXCHG ~Ev,~Gv",   "LSS ~Mp",           "BTR ~Ev,~Gv",
  "LFS ~Mp",           "LGS ~Mp",           "MOVZX ~Gv,~Eb",     "MOVZX ~Gv,~Ew",
  0,                   0,                   "~g8 ~Ev,~Ib",       "BTC ~Ev,~Gv",
  "BSF ~Gv,~Ev",       "BSR ~Gv,~Ev",       "MOVSX ~Gv,~Eb",     "MOVSX ~Gv,~Ew",

/* Cx */
  "XADD ~Eb,~Gb",      "XADD ~Ev,~Gv",      0,                   0,
  0,                   0,                   0,                   "CMPXCHG8B ~Ad",
  "BSWAP EAX",         "BSWAP ECX",         "BSWAP EDX",         "BSWAP EBX",
  "BSWAP ESP",         "BSWAP EBP",         "BSWAP ESI",         "BSWAP EDI",

/* Dx */
  0,                   "PSRLW ~GM,~Eq",     "PSRLD ~GM,~Eq",     "PSRLQ ~GM,~Eq",
  0,                   "PMULLW ~GM,~Eq",    0,                   0,
  "PSUBUSB ~GM,~Eq",   "PSUBUSW ~GM,~Eq",   0,                   "PAND ~Eq",
  "PADDUSB ~Eq",       "PADDUSW ~Eq",       0,                   "PANDN ~Eq",

/* Ex */
  0,                   "PSRAW ~GM,~Eq",     "PSRAD ~GM,~Eq",     0,
  0,                   "PMULHW ~GM,~Eq",    0,                   0,
  "PSUBSB ~GM,~Eq",    "PSUBSW ~GM,~Eq",    0,                   "POR ~GM,~Eq",
  "PADDSB ~Eq",        "PADDSW ~Eq",        0,                   "PXOR ~GM,~Eq",

/* Fx */
  0,                   "PSLLW ~GM,~Eq",     "PSLLD ~GM,~Eq",     "PSLLQ ~GM,~Eq",
  0,                   "PMADDWD ~GM,~Eq",   0,                   "MASKMOVQ ~GM,~GM",
  "PSUBB ~GM,~Eq",     "PSUBW ~GM,~Eq",     "PSUBD ~GM,~Eq",     0,
  "PADDB ~GM,~Eq",     "PADDW ~GM,~Eq",     "PADDD ~GM,~Eq",     "OIO"
};


static char *groups[10][8] = {   /* group 0 is group 3 for ~Ev set */

/* 0 */ { "TEST ~Ev,~Iv",  "TEST ~Ev,~Iv,",  "NOT ~Ev",      "NEG ~Ev",
          "MUL ~eAX,~Ev",  "IMUL ~eAX,~Ev",  "DIV ~Ev",      "IDIV ~Ev"    },
/* 1 */ { "ADD",           "OR",             "ADC",          "SBB",
          "AND",           "SUB",            "XOR",          "CMP"         },
/* 2 */ { "ROL",           "ROR",            "RCL",          "RCR",
          "SHL",           "SHR",            "SHL",          "SAR"         },
/* 3 */ { "TEST ~Eb,~Ib",  "TEST ~Eb,~Ib,",  "NOT ~Eb",      "NEG ~Eb",
          "MUL AL,~Eb",    "IMUL AL,~Eb",    "DIV AL,~Eb",   "IDIV AL,~Eb" },
/* 4 */ { "INC ~Eb",       "DEC ~Eb",        0,              0,
          0,               0,                0,              0             },
/* 5 */ { "INC ~Ev",       "DEC ~Ev",        "CALL ~Ev",     "CALL ~Ep",
          "JMP ~Ev",       "JMP ~Ep",        "PUSH ~Ev",     0             },
/* 6 */ { "SLDT ~Ew",      "STR ~Ew",        "LLDT ~Ew",     "LTR ~Ew",
          "VERR ~Ew",      "VERW ~Ew",       0,              0             },
/* 7 */ { "SGDT ~Ms",      "SIDT ~Ms",       "LGDT ~Ms",     "LIDT ~Ms",
          "SMSW ~Ew",      0,                "LMSW ~Ew",     "INVLPG ~Av"  },
/* 8 */ { 0,               0,                0,              0,
          "BT",            "BTS",            "BTR",          "BTC"         },
/* 9 */ { "PSRAW",         0,                "PSRLD",        0,
          "PSRAD",         0,                "PSLLW",        "PSRLQ"       }
     };

/* for display
 */
static char *seg_names [] = { "ES", "CS", "SS", "DS", "FS", "GS"              };
static char *breg_names[] = { "AL", "CL", "DL", "BL", "AH", "CH", "DH", "BH"  };
static char *wreg_names[] = { "AX", "CX", "DX", "BX", "SP", "BP", "SI", "DI"  };
static char *dreg_names[] = { "EAX","ECX","EDX","EBX","ESP","EBP","ESI","EDI" };

/* zero here means invalid.  If first entry starts with '*', use ST(i)
 * no assumed ~EFs here.  Indexed by rm(ModRM())
 */
static char *fop_0[8]  = { 0,         0, 0, 0, 0, 0, 0, 0             };
static char *fop_9[8]  = { "*FXCH ~GF"                                };
static char *fop_10[8] = { "FNOP",    0, 0, 0, 0, 0, 0, 0             };
static char *fop_12[8] = { "FCHS",    "FABS",    0,         0,
                           "FTST",    "FXAM",    0,         0         };
static char *fop_13[8] = { "FLD1",    "FLDL2T",  "FLDL2E",  "FLDPI",
                           "FLDLG2",  "FLDLN2",  "FLDZ",    0         };
static char *fop_14[8] = { "F2XM1",   "FYL2X",   "FPTAN",   "FPATAN",
                           "FXTRACT", "FPREM1",  "FDECSTP", "FINCSTP" };
static char *fop_15[8] = { "FPREM",   "FYL2XP1", "FSQRT",   "FSINCOS",
                           "FRNDINT", "FSCALE",  "FSIN",    "FCOS"    };
static char *fop_21[8] = { 0,         "FUCOMPP", 0,         0,
                           0,         0,         0,         0         };
static char *fop_28[8] = { 0,         0,         "FCLEX",   "FINIT",
                           0,         0,         0,         0         };

static char *fop_32[] = { "*FADD ~GF,ST(0)"   };
static char *fop_33[] = { "*FMUL ~GF,ST(0)"   };
static char *fop_36[] = { "*FSUBR ~GF,ST(0)"  };
static char *fop_37[] = { "*FSUB ~GF,ST(0)"   };
static char *fop_38[] = { "*FDIVR ~GF,ST(0)"  };
static char *fop_39[] = { "*FDIV ~GF,ST(0)"   };
static char *fop_40[] = { "*FFREE ~GF"        };
static char *fop_42[] = { "*FST ~GF"          };
static char *fop_43[] = { "*FSTP ~GF"         };
static char *fop_44[] = { "*FUCOM ~GF"        };
static char *fop_45[] = { "*FUCOMP ~GF"       };
static char *fop_48[] = { "*FADDP ~GF,ST(0)"  };
static char *fop_49[] = { "*FMULP ~GF,ST(0)"  };
static char *fop_51[] = { 0, "FCOMPP", 0, 0, 0, 0, 0, 0 };
static char *fop_52[] = { "*FSUBRP ~GF,ST(0)" };
static char *fop_53[] = { "*FSUBP ~GF,ST(0)"  };
static char *fop_54[] = { "*FDIVRP ~GF,ST(0)" };
static char *fop_55[] = { "*FDIVP ~GF,ST(0)"  };
static char *fop_60[] = { "FSTSW AX", 0, 0, 0, 0, 0, 0, 0 };

static char **fspecial [8*8] = { /* NULL = use ST(i) */
            NULL,   NULL,   NULL,   NULL,   NULL,   NULL,   NULL,   NULL,
            NULL,   fop_9,  fop_10, NULL,   fop_12, fop_13, fop_14, fop_15,
            fop_0,  fop_0,  fop_0,  fop_0,  fop_0,  fop_21, fop_0,  fop_0,
            fop_0,  fop_0,  fop_0,  fop_0,  fop_28, fop_0,  fop_0,  fop_0,
            fop_32, fop_33, fop_0,  fop_0,  fop_36, fop_37, fop_38, fop_39,
            fop_40, fop_0,  fop_42, fop_43, fop_44, fop_45, fop_0,  fop_0,
            fop_48, fop_49, fop_0,  fop_51, fop_52, fop_53, fop_54, fop_55,
            fop_0,  fop_0,  fop_0,  fop_0,  fop_60, fop_0,  fop_0,  fop_0
          };

/* Assumed " ~EF" at end of each. mod != 3 only
 */
static char *floatops [4*16] = { 
     /* 0x */ "FADD",    "FMUL",    "FCOM",   "FCOMP",
              "FSUB",    "FSUBR",   "FDIV",   "FDIVR",
              "FLD",     0,         "FST",    "FSTP",
              "FLDENV",  "FLDCW",   "FSTENV", "FSTCW",
     /* 1x */ "FIADD",   "FIMUL",   "FICOMW", "FICOMPW",
              "FISUB",   "FISUBR",  "FIDIV",  "FIDIVR",
              "FILD",    0,         "FIST",   "FISTP",
              "FRSTOR",  "FLDT",    0,        "FSTPT",
     /* 2x */ "FADDQ",   "FMULQ",   "FCOMQ",  "FCOMPQ",
              "FSUBQ",   "FSUBRQ",  "FDIVQ",  "FDIVRQ",
              "FLDQ",    0,         "FSTQ",   "FSTPQ",
              0,         0,         "FSAVE",  "FSTSWW",
     /* 3x */ "FIADDW",  "FIMULW",  "FICOMW", "FICOMPW",
              "FISUBW",  "FISUBRW", "FIDIVW", "FIDIVR",
              "FILDW",   0,         "FISTW",  "FISTPW",
              "FBLDT",   "FILDQ",   "FBSTPT", "FISTPQ"
            };

static int   prefix, opsize, addrsize, modrmv, sibv;
static UCHAR *addrIn, opCode;

static char  outBuf [MAX_SYM_NAME];
static char *outPtr = outBuf;
static char *endPtr = outBuf + MAX_SYM_NAME - 1;

#ifdef __GNUC__
static void outf (char *fmt, ...) __attribute__ ((format(printf,1,2)));
#endif

static void outf (char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);
  outPtr += vsprintk (outPtr, fmt, args);
  va_end (args);
}

#define OUTC(c)  do {                   \
                   if (outPtr < endPtr) \
                      *outPtr++ = c;    \
                 } while (0)

#define OUTS(s)  do { char *p = s; \
                    while (*p)     \
                      OUTC(*p++);  \
                 } while (0)

#define B_PTR   "byte ptr"
#define W_PTR   "word ptr"
#define DW_PTR  "dword ptr"
#define PW_PTR  "pword ptr"

/*
 * Gets a byte to disassemble and update addrIn.
 */
static UCHAR GetByte (void)
{
  return (opCode = *addrIn++);
}

static void OutSymName (ULONG eip)
{
  char *sym = GetSymbolStr (eip);

  if (*sym)
       OUTS (sym);
  else outf ("%lX",eip);
}

/*
 * Get Mod/RM field byte for current instruction
 */
static int ModRM (void)
{
  if (modrmv == -1)
      modrmv = GetByte();
  return (modrmv);
}


/*
 * Get 'scale-index-base' byte for current instruction
 */
static int sib (void)
{
  if (sibv == -1)
      sibv = GetByte();
  return (sibv);
}

/*
 * The register is encoded as bit 3,4,5 in the byte.
 * xxRRRxxx. This macro extracts it.
 */
#define REG(a)  (((a) >> 3) & 7)
#define MOD(a)  (((a) >> 6) & 7)
#define RM(a)   ((a) & 7)

/*
 * Determines how many bytes left in the instruction from the
 * letter in the table (which is passed in here).
 */
static int Bytes (char c)
{
  switch (c)
  {
    case 'b': return (1);
    case 'w': return (2);
    case 'd': return (4);
    case 'v': return (opsize == 32 ? 4 : 2);
  }
  return (0);
}

static char * AsmHex (UCHAR b)
{
  static char buf[10];
  if (b >= 0xA0)
       sprintk (buf,"%03Xh",b);
  else sprintk (buf,"%02Xh",b);
  return (buf);
}

#if 0
static void ohex_ptr (char c, int defsize)
{
  int n = 1;
  static char *ptr_name[] = { "BYTE","WORD","","DWORD","","PWORD" };
  switch (c)
  {
    case 'b': n = 1; break;
    case 'w': n = 2; break;
    case 'd': n = 4; break;
    case 's': n = 6; break;
    case 'c':
    case 'v': n = (defsize == 32) ? 4 : 2; break;
    case 'p': n = (defsize == 32) ? 6 : 4; break;  /* 32 or 48 bit pointer */
  }
  outf ("%s PTR ",ptr_name[n-1]);
}
#endif

/*
 * Get the correct number of bytes for immediate data from the
 * code stream and output it as hex.
 */

#define CTRL_IMM  1
#define CTRL_PLUS 2

static void ohex (char ctrl, char c, int extend, int optional, int defsize)
{
  UCHAR  buf[8];
  int    i, n = 0, s = 0;
  long  *x   = (long *)&buf;
  ULONG *ofs = (ULONG*)&buf;
 
  switch (c)
  {
    case 'a': break;
    case 'b': n = 1;   /* byte  */
              break;
    case 'w': n = 2;   /* word  */
              break;
    case 'd': n = 4;   /* dword */
              break;
    case 's': n = 6;   /* fword */
              break;
    case 'q': n = 8;   /* qword */
              break;
    case 'c':
    case 'v': n = (defsize == 32) ? 4 : 2;
              break;
    case 'p': n = (defsize == 32) ? 6 : 4;  /* 32 or 48 bit pointer */
              s = 1;
              break;
  }

  for (i = 0; i < n; i++)
      buf[i] = GetByte();

  if ((c == 'v' || c == 'd') && n == 4)
  {
    SymNode *sp = FindPublicSym (*ofs);

    if (sp && sp->ofs == *ofs)
    {
      outf ("%s%s", (ctrl & CTRL_IMM) ? "offset " : "", sp->name);
      return;
    }
  }

  /* sign extend the value into a dword
   */
  for (; i < extend; i++)
      buf[i] = (buf[i-1] & 0x80) ? 0xFF : 0;

  if (n <= 4 && extend && *x < 0)
  {
    outf ("-%02lX", -*x);
    return;
  }
  if (ctrl & CTRL_PLUS)
     OUTC ('+');

  if (s)       /* outputs the segment value of FAR pointer */
  {
    outf ("%02X%02X:",buf[n-1],buf[n-2]);
    n -= 2;
  }

  if (extend > n)
  {
    if (*ofs && !optional)
       OUTC ('+');
    n = 4;
  }

  switch (n)
  {
    case 1: if (*ofs & 0xFF)
               outf ("%02X", (UCHAR)(*ofs & 0xFF));
            break;
    case 2: if (*ofs & 0xFFFF)
               outf ("%04X", (USHORT)(*ofs & 0xFFFF));
            break;
    case 4: if (*ofs)
               outf ("%08lX", *ofs);
            break;
    case 8: if (*(ULONG*)&buf[4] || *(ULONG*)&buf[0])
               outf ("%08lX%08lX", *(ULONG*)&buf[4], *(ULONG*)&buf[0]);
            break;

  }
}

/*------------------------------------------------------------------------*/

static void RegName (UCHAR which, char size)
{
  if (size == 'F')
  {
    outf ("ST(%d)",which);
    return;
  }

  if (((size == 'v') && (opsize == 32)) || (size == 'd'))
     OUTC ('E');

  if (size == 'b')
       OUTS (breg_names[which]);
  else OUTS (wreg_names[which]);
}

/*------------------------------------------------------------------------*/

static void floating_point (int e1)
{
  UCHAR b2  = ModRM();
  UCHAR esc = 8 * e1 + REG(b2);

  if (MOD(b2) == 3)
  {
    if (fspecial[esc])
    {
      if (fspecial[esc][0])
      {
        if (fspecial[esc][0][0] == '*')
             Decode (fspecial[esc][0]+1);
        else Decode (fspecial[esc][RM(b2)]);
      }
      else
      {
        outf ("DB %s,", AsmHex(e1+0xD8));
        outf ("%s ;illegal float op-code", AsmHex(b2));
      }
    }
    else
    {
      Decode (floatops[esc]);
      Decode (" ~EF");
    }
  }
  else
  {
    Decode (floatops[esc]);
    Decode (" ~EF");
  }
}


/*
 *  This takes in two chars that represent part of the op code and
 *  puts out the proper text to match what the letter represents.
 *  c is the first char after the tilde and t is next one. See
 *  opcode1[] strings for what the chars mean.
 */
static void escape (char c, char t)
{
  long  delta, vals = 0;
  UCHAR b2;
  int   index;
  int   extend  = (addrsize == 32) ? 4 : 2;
  char  ip2name = 0;

  switch (c)
  {
    case 'A':                             /* Direct Address */
         ohex (CTRL_IMM,t,extend,0,addrsize);
         break;

    case 'C':                             /* Reg of R/M picks control reg */
         outf ("CR%d",REG(ModRM()));
         break;

    case 'D':                             /* Reg of R/M pick debug reg */
         outf ("DR%d",ModRM());
         break;

    case 'E':                             /* R/M picks operand */
         do_ModRM (t);
         break;

    case 'G':                             /* Reg of R/M picks general reg */
         switch (t)
         {
           case 'F': RegName (ModRM() & 7, 'F');
                     break;
           case 'M': outf ("MM%d", REG(ModRM()));
                     break;
           case 'X': outf ("XMM%d", REG(ModRM()));
                     break;
           default : RegName (REG(ModRM()), t);
         }
         break;

    case 'I':                             /* Immediate data */
         ohex (CTRL_IMM,t,0,0,opsize);
         break;

    case 'j': ip2name = 1;
    case 'J':                             /* Relative IP offset */
         switch (Bytes(t))
         {
           case 1:
                vals = GetByte();         /* must remain signed! */
                break;
           case 2:
                vals  = GetByte();        /* made SIGNEd bytes/Words */
                vals += GetByte() << 8;
                break;
           case 4:
                vals  = GetByte();
                vals |= GetByte() << 8;
                vals |= GetByte() << 16;
                vals |= GetByte() << 24;
                break;
           default:
                vals = 0;
         }
         delta = (long) ((signed)addrIn + vals);
         if (ip2name)
              OutSymName (delta);
         else outf ("%lX",delta);
         break;

    case 'M':                             /* R/M picks memory */
         do_ModRM (t);
         break;

    case 'O':                             /* NO R/M, Offset only */
         Decode ("~p:[");
         ohex (0,t,extend,0,addrsize);
         OUTC (']');
         break;

    case 'R':                             /* Mod of R/M pick REG only */
         do_ModRM (t);
         break;

    case 'S':                             /* Reg of R/M picks seg reg */
         OUTS (seg_names[REG(ModRM())]);
         break;

    case 'T':                             /* Reg of R/M picks test reg */
         outf ("TR%d",ModRM());
         break;

    case 'X':                             /* DS:(E)SI */
         if (opsize == 32)
              OUTS ("[ESI]");
         else OUTS ("[SI]");
         break;

    case 'Y':                             /* ES:EDI */
         if (opsize == 32)
              OUTS ("[EDI]");
         else OUTS ("[DI]");
         break;

    case 'Z':                             /* Sink for AAD/AAM */
         b2 = GetByte();
         if (b2 != 10)
            outf ("%u",b2);
         break;

    case '2':                             /* Prefix of 2 byte opcode */
         b2 = GetByte();
         if (secOpcode[b2])
              Decode (secOpcode[b2]);
         else outf ("DB %s ;illegal 2-byte op-code (0F xx)", AsmHex(b2));
         break;

    case 'e':                   /* t is part of reg name */
         if (opsize == 32)
         {
           if (t == 'W')        /* put out "D" if t is "W" on 32 bit opsize */
                OUTC ('D');
           else outf ("E%c",t); /* put out "E". if t <> "W" then put t */
         }
         else
           OUTC (t);
         break;

    case 'f':
         floating_point (t-'0');
         break;

    case 'g':                             /* do R/M group 'n' */
         if (t >= 'A')
              index = t - 'A' + 10;
         else index = t - '0';
         Decode (groups[index][REG(ModRM())]);
         break;

    case 'p':                             /* Segment prefix */
         switch (t)
         {
           case 'C':                      /* CS */
           case 'D':                      /* DS */
           case 'E':                      /* ES */
           case 'F':                      /* FS */
           case 'G':                      /* GS */
           case 'S':                      /* SS */
                prefix = t;
           case '*':
                Decode (opmap1[GetByte()]);
                break;
           case ':':
                if (prefix)
                   outf ("%cS:",prefix);
                break;
         }
         break;

    case 's':                          /* Size override  */
         switch (t)
         {
           case 'a':
                addrsize = 48 - addrsize;
                Decode (opmap1[GetByte()]);
                break;
           case 'o':
                opsize = 48 - opsize;
                Decode (opmap1[GetByte()]);
                break;
         }
         break;

    case 'x':
         if (opsize == 32)     /* use32: J~xCXZ -> JECXZ */
            OUTC ('E');        /* use16: J~xCXZ -> JCXZ  */
  }
}


/*
 *  This expands and outputs the instruction
 *  string passed in if it finds the escape
 *  character (tilde).
 */ 
static void Decode (char *s)
{
  char *s_end;

  if (!s)                  /* undefined opcode */
  {
    outf ("DB %s ;illegal op-code", AsmHex(opCode));
    return;
  }

  for (s_end = s + strlen(s); *s && s < s_end; s++)
  {
    char c = *s;

    if (c == '~')         /* if it's a `~' then ESCAPE */
    {
      c = *(++s);
      escape (c,*(++s));
    }
    else
      OUTC (c);           /* else put out the char found! */
  }
}           

/*
 * outputs 'scale-index-base' instructions
 */ 
static void do_sib (int m)
{
  int s = ((sib()) >> 6) & 7;            /* SSxxxxxx Scale */
  int i = ((sib()) >> 3) & 7;            /* xxIIIxxx Index */
  int b = sib() & 7;                     /* xxxxxBBB Base  */

  switch (b)
  {
    case 0: Decode ("~p:[EAX"); break;
    case 1: Decode ("~p:[ECX"); break;
    case 2: Decode ("~p:[EDX"); break;
    case 3: Decode ("~p:[EBX"); break;
    case 4: Decode ("~p:[ESP"); break;
    case 6: Decode ("~p:[ESI"); break;
    case 7: Decode ("~p:[EDI"); break;
    case 5: if (m == 0)
            {
              Decode ("~p:[");
              ohex (0,'d',4,0,addrsize);
            }
            else Decode ("~p:[EBP");
            break;
  }

  if (i != 4)
  {
    outf ("+%s", dreg_names[i]);
    if (s)
       outf ("*%d", 1 << s);
  }
}

/*------------------------------------------------------------------------*/

static void do_ModRM (char t)
{
  int m = ((ModRM()) >> 6) & 7;
  int r = ModRM() & 7;
  int extend = (addrsize == 32) ? 4 : 2;

  if (m == 3)
  {
    RegName (r,t);
    return;
  }

  if ((m == 0) && (r == 5) && (addrsize == 32))
  {
    Decode ("~p:[");
    ohex (0,'d',extend,0,addrsize);
    OUTC (']');
    return;
  }

  if ((m == 0) && (r == 6) && (addrsize == 16))
  {
    Decode ("~p:[");
    ohex (0,'w',extend,0,addrsize);
    OUTC (']');
    return;
  }

  if (r == 4)
     do_sib (m);
  else
  {
    Decode ("~p:[");
    OUTS (dreg_names[r]);
  }

  switch (m)
  {
    case 1: ohex (0,'b',extend,0,addrsize);
            break;
    case 2: ohex (CTRL_PLUS,'v',extend,0,addrsize);
            break;
  }
  OUTC (']');
}

/*
 * This disassembles one instruction each time it
 * is called. `addr' is updated to point to next instruction.
 * Return value is disassembly string.
 */
const char * Disassemble (void ** addr)
{
  prefix = 0;
  modrmv = sibv = -1;
  opsize = addrsize = 8 * sizeof(int); /* operand size is 16/32 bit */
  addrIn = *addr;
  outPtr = outBuf;

  Decode (opmap1[GetByte()]);          /* decode instruction and output */
  *addr   = (void*)addrIn;
  *outPtr = 0;
  return (const char*) &outBuf[0];
}
